/*
 * ECRTS 2010 submission "Sulphur"
 * SMMU model
 * Copyright (C) Jack Whitham 2010
 */

#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include "smmu.h"


typedef struct SMMU_Table_Entry_struct {
    int                 is_dirty, is_free, is_read_only;
    ulongint            base, limit, size, offset, id;
} SMMU_Table_Entry;

typedef struct SMMU_Params_Struct {
    Tetra_Params *      tp;
    SMMU_Table_Entry *  table;
    ulongint *          scratchpad;
    SMMU_Close_Mode     close_mode;
    ulongint            table_size;
    ulongint            space;
} SMMU_Params;



static int SMMU_Worth_Copying (void * cb, ulongint size, 
        ulongint access_count, int read_only)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    Tetra_Params *  tp = sp->tp;
    ulongint        direct_cost = (tp->buslat + 1) * access_count;
    ulongint        smmu_cost = tp->calc_txfer_fn(tp, size, 0);

    if (!read_only) smmu_cost *= 2;
    smmu_cost += access_count;

    if (direct_cost > smmu_cost) {
        ulongint score = direct_cost - smmu_cost;
        if (score > INT_MAX) {
            score = INT_MAX;
        }
        return (int) score;
    }
    return 0;
}

static ulongint SMMU_Close (void * cb, ulongint value)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    Tetra_Params *  tp = sp->tp;
    ulongint        n, i, j;
    SMMU_Table_Entry * te;

    /* Locate table entry */
    n = value >> PTR_ID_OFFSET;
    assert (1 <= n);
    assert (n <= sp->table_size);
    te = &sp->table[ n - 1 ];
    assert (!te->is_free);
    assert (te->id == n);
    
    /* Ensure that the scratchpad space is marked as free */
    assert ((te->offset + te->size) <= sp->space);
    for (i = 0; i < te->size; i++) {
        j = i + te->offset;
        assert (sp->scratchpad[ j ] == n);
        sp->scratchpad[ j ] = 0;
    }
    te->is_free = 1;

    assert ((sp->close_mode == CM_ORIGINAL) ||
            (sp->close_mode == CM_DYNAMIC_IF_DIRTY) ||
            (sp->close_mode == CM_OPTION_B) ||
            (sp->close_mode == CM_OPTION_C));

    /* Do close writeback under certain circumstances */
    if ((sp->close_mode == CM_ORIGINAL) 
    || ((sp->close_mode == CM_DYNAMIC_IF_DIRTY) && te->is_dirty)
    || (((sp->close_mode == CM_OPTION_B) || (sp->close_mode == CM_OPTION_C))
    && ! te->is_read_only)) {
        ulongint extra_writes = 0;

        if (sp->close_mode == CM_OPTION_C) {
            extra_writes = te->size * (sp->table_size - 1);
        }
        tp->do_txfer_fn(tp, te->size, extra_writes);
    }
    return value & ADDRESS_MASK;
}

static ulongint SMMU_Open (void * cb, ulongint value,
        ulongint size, ulongint * offset, int read_only)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    Tetra_Params *  tp = sp->tp;
    ulongint        i, j, k, orig_value;
    SMMU_Table_Entry * te;

    value &= ADDRESS_MASK;
    orig_value = value;

    if (tp->smmu_powers_of_two) {
        /* This is used to save space & time in the address matching 
         * logic. It's much easier to match with bit masks than with
         * comparators. BUT AT WHAT COST!?
         * 1. Round up size to nearest 2^x */
        ulongint rnd_size = 0;
        ulongint rnd_bits, mask;
        for (rnd_bits = 0; rnd_bits < 32; rnd_bits++) {
            rnd_size = 1 << rnd_bits;
            if (size <= rnd_size) break;
        }
        /* 2. Round up once more to deal with arbitrary address */
        rnd_bits ++;
        size = 1 << rnd_bits;
        /* 3. Round down address accordingly */
        mask = (~0) << rnd_bits;
        value &= mask;
        assert (value <= orig_value);
        assert (orig_value < (value + size));
        assert (size < MEM_SIZE);
    }

    if ((size > sp->space)
    || ((* offset) >= sp->space)
    || (((* offset) + size) > sp->space)) {
        /* Can't OPEN that! Go away. */
        return orig_value;
    }

    /* Ensure that the scratchpad space is marked as free */
    for (i = 0; i < size; i++) {
        j = i + (* offset);
        assert (sp->scratchpad[ j ] == 0);
    }

    for (i = 0; i < sp->table_size; i++) {
        te = &sp->table[ i ];
        if (!te->is_free) continue;

        /* Free table entry found */
        assert (te->id == (i + 1));
        te->is_free = 0;
        te->base = value;
        te->limit = value + size;
        te->offset = (* offset);
        te->size = size;
        te->is_read_only = read_only && 
            ((sp->close_mode == CM_OPTION_B) ||
            (sp->close_mode == CM_OPTION_C));

        /* Add cost of OPEN */
        tp->do_txfer_fn(tp, size, 0);

        /* Sanity check number of RO entries */
        if (sp->close_mode == CM_OPTION_B) {
            ulongint ro_count = 0;
            for (j = 0; j < sp->table_size; j++) {
                SMMU_Table_Entry * te2 = &sp->table[ j ];
                if (te2->is_read_only && ! te2->is_free) {
                    ro_count ++;
                }
            }
            assert (ro_count <= tp->ro_max);
        }

        /* Allocate space */
        for (k = 0; k < size; k++) {
            j = k + (* offset);
            assert (sp->scratchpad[ j ] == 0);
            sp->scratchpad[ j ] = te->id;
        }

        (* offset) += size;
        return (te->id << PTR_ID_OFFSET) | orig_value;
    }
    assert (0); /* Table full */
    return orig_value;
}


static void Access (void * cb, ulongint value, const char * comment, int write)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    Tetra_Params *  tp = sp->tp;
    ulongint        i;
    SMMU_Table_Entry * te;
    ulongint        address = value & ADDRESS_MASK;
    int             table_hit = 0;
    int             force_off_chip = 0;
    
    if ((tp->smmu_very_deterministic)
    && ((value >> PTR_ID_OFFSET) == 0)) {
        /* shouldn't be a hit.. */
        force_off_chip = 1;
    }

    /* Resolve for RW table entries */
    for (i = 0; i < sp->table_size; i++) {
        te = &sp->table[ i ];

        if ((te->base <= address) && (address < te->limit)
        && !te->is_read_only) {
            if (write) {
                te->is_dirty = 1;
            }
            table_hit = 1;
            break;
        }
    }
    /* Resolve for RO table entries */
    if (!table_hit) {
        for (i = 0; i < sp->table_size; i++) {
            te = &sp->table[ i ];

            if ((te->base <= address) && (address < te->limit)
            && te->is_read_only) {
                assert ((sp->close_mode == CM_OPTION_B)
                    || (sp->close_mode == CM_OPTION_C));
                if (write) {
                    /* Read-only entry hit with a write!
                     * Force update to main memory */
                    te->is_dirty = 1;
                    force_off_chip = 1;
                }
                table_hit = 1;
                break;
            }
        }
    }
    if (!table_hit) {
        /* No table entry? */
        force_off_chip = 1;
    }

    if (force_off_chip) {
        tp->do_txfer_fn(tp, 1, 0);
    } else {
        tp->do_tick_fn(tp, TICK_ON_CHIP_ACCESS);
    }
}

static void SMMU_Read (void * cb, ulongint value, const char * comment)
{
    Access(cb, value, comment, 0);
}

static void SMMU_Write (void * cb, ulongint value, const char * comment)
{
    Access(cb, value, comment, 1);
}

static void SMMU_Destroy (void * cb)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    Tetra_Params *  tp = sp->tp;

    free(sp->scratchpad);
    free(sp->table);
    free(sp);
    tp->mem_model_cb = NULL;
}

static void SMMU_Flush (void * cb)
{
    SMMU_Params *   sp = (SMMU_Params *) cb;
    ulongint        i;

    for (i = 0; i < sp->table_size; i ++) {
        if (!sp->table[ i ].is_free) {
            SMMU_Close(cb, (i + 1) << PTR_ID_OFFSET);
        }
    }
}

void Init_SMMU(Tetra_Params * tp, 
        ulongint table_size, SMMU_Close_Mode close_mode)
{
    SMMU_Params *   sp = calloc(1, sizeof(SMMU_Params));
    ulongint        i;

    assert (table_size > 0);
    sp->tp = tp;
    sp->space = tp->bytes_on_chip;
    sp->table_size = table_size;
    sp->close_mode = close_mode;
    sp->scratchpad = calloc(sp->space, sizeof(ulongint));
    sp->table = calloc(sp->table_size, sizeof(SMMU_Table_Entry));

    tp->close_fn = SMMU_Close;
    tp->open_fn = SMMU_Open;
    tp->read_fn = SMMU_Read;
    tp->write_fn = SMMU_Write;
    tp->worth_copying_fn = SMMU_Worth_Copying;
    tp->destroy_fn = SMMU_Destroy;
    tp->flush_fn = SMMU_Flush;
    tp->mem_model_cb = sp;

    for (i = 0; i < sp->table_size; i ++) {
        sp->table[ i ].is_free = 1;
        sp->table[ i ].id = i + 1;
    }
    SMMU_Flush(sp);
}

