
# What's the point?
# 1. Classify memory operations into groups:
#        * Random       - any access
#        * Linear       - access sequence that steps through memory, e.g. foo[i]
#        * LinearStruct - a cluster of fields in a struct, e.g. foo[i].j
#        * Local        - single access to dynamic location
#        * Static       - single access to static location
# 2. Plot relative frequencies of each type in each program
# 

import os, sys, pickle, random

TILE_SIZE = 128
MAX_LOOP_SIZE = 20000
ROUNDUP = 7
PROGRAMS = dict()
ALIGN = True
rng = random.Random(1)

def Pre_Proc(name):
    if len(name) > 12:
        name = name[ : 9 ] + "..."
    name = name.replace('%', '\\%').replace('_', '\\_')
    name = name.rjust(15)
    return name

def Make_Debug_Dir(name):
    max_length = 4
    debug_dir = "benchmarks/debug/%s-%u" % (name, max_length)

    try:
        os.mkdir(debug_dir)
    except:
        pass

    return debug_dir

def Get_Sulphur(name):
    if PROGRAMS.has_key(name):
        return PROGRAMS[ name ]

    debug_dir = os.path.abspath(Make_Debug_Dir(name))
    sulphur_dat = "%s/sulphur.dat" % debug_dir
    try:
        sulphur = pickle.load(file(sulphur_dat, "rb"))
    except:
        sulphur = None
    func_for_region_dat = "%s/func_for_region.dat" % debug_dir
    try:
        func_for_region = pickle.load(file(func_for_region_dat, "rb"))
    except:
        func_for_region = None

    for rdata in sulphur:
        rn = rdata[ 'region_num' ]
        rdata[ 'func_name' ] = func_for_region.get(rn, "r%d" % rn)

    PROGRAMS[ name ] = sulphur
    return sulphur

def Step(step):
    if step < 0:
        return '%d' % step
    else:
        return '+%d' % step

class AccessType:
    def __init__(self, name, count, store_count):
        self.name = name
        self.count = count
        self.store_count = store_count

    def Read_Only(self):
        return self.store_count == 0

    def Info(self):
        return ""

    def Get_Message(self):
        return ""

    def Get_Sequence_Info(self):
        return []

    def Get_Abs_Bound(self, max_loop_size):
        return 1 << 16

    def Get_Page_Size(self):
        return 1

    def Get_Size_Equation(self):
        return "1"

    def Init_Random(self):
        return False

    def Get_Align(self):
        return False

    def Pattern(self):
        return ""

class Linear(AccessType):
    def __init__(self, name, count, store_count, max_factor, step = 1):
        AccessType.__init__(self, name, count, store_count)
        assert step != 0
        self.step = step
        self.max_factor = max_factor
                
    def Pattern(self):
        return "S"

    def Info(self):
        return Step(self.step)

    def Get_Message(self):
        return "%s: Linear, count %d, step %d" % (
                    self.name, self.count, self.step)

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("(i * %d)" % self.step, write)
            
    def Get_Page_Size(self):
        return abs(self.step)

    def Get_Size_Equation(self):
        return "(loop_size * %d)" % abs(self.step)

    def Get_Abs_Bound(self, max_loop_size):
        return abs(max_loop_size * self.step)


class Unrolled(AccessType):
    def __init__(self, name, count, store_count, max_factor, step = 1):
        AccessType.__init__(self, name, count, store_count)
        assert step != 0
        self.step = step 
        self.max_factor = max_factor
                
    def Pattern(self):
        return "D"

    def Info(self):
        return ""

    def Get_Message(self):
        return "%s: Unrolled, count %d, step %d" % (
                    self.name, self.count, self.step)

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("%d" % (i * self.step), write)
            
    def Get_Page_Size(self):
        return abs(self.step)

    def Get_Size_Equation(self):
        return "(%d)" % abs(self.step * self.count)

    def Init_Random(self):
        return True

    def Get_Abs_Bound(self, max_loop_size):
        return abs(self.step * self.count)

class LinearStruct(AccessType):
    def __init__(self, name, count, store_count,
                    max_factor, step = 1, size = 1):
        AccessType.__init__(self, name, count, store_count)
        assert step != 0
        self.step = step
        self.size = size
        self.max_factor = max_factor

    def Pattern(self):
        return "S"

    def Info(self):
        return Step(self.step)

    def Get_Message(self):
        return "%s: Struct, count %d, step %d" % (
                    self.name, self.count, self.step)

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("((i * %d) + %d)" % (self.step, 
                    rng.randint(0, self.step - 1)), write)
            
    def Get_Page_Size(self):
        return abs(self.step)

    def Get_Size_Equation(self):
        return "(loop_size * %d)" % abs(self.step)

    def Get_Abs_Bound(self, max_loop_size):
        return abs(max_loop_size * self.step)

    def Get_Align(self):
        return True # Yes, array begins with a pointer

class Local(AccessType):
    def __init__(self, name, count, store_count):
        AccessType.__init__(self, name, count, store_count)

    def Pattern(self):
        return "C"

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("0", write)

    def Get_Message(self):
        return "%s: Local, count %d" % (self.name, self.count)

    def Get_Page_Size(self):
        return 1

    def Get_Size_Equation(self):
        return "1"

class Static(AccessType):
    def __init__(self, name, count, store_count):
        AccessType.__init__(self, name, count, store_count)

    def Pattern(self):
        return "C"

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("0", write)

    def Get_Message(self):
        return "%s: Static, count %d" % (self.name, self.count)

    def Get_Page_Size(self):
        return 1

    def Get_Size_Equation(self):
        return "1"

class Literal(AccessType):
    def __init__(self, count):
        AccessType.__init__(self, "huh", count, 0)

    def Pattern(self):
        return "C"

    def Get_Message(self):
        return "%s: Literal" % self.name

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            yield ("0", False)

    def Get_Page_Size(self):
        return 1

    def Get_Size_Equation(self):
        return "1"

class Random(AccessType):
    def __init__(self, name, count, store_count, size = None):
        AccessType.__init__(self, name, count, store_count)
        if size == None:
            size = 1 << 20
        assert size > 0
        self.size = size

    def Pattern(self):
        return "R"

    def Info(self):
        M = float(1 << 20)
        k = float(1 << 10)
        if self.size >= M:
            return '[%1.1fM]' % (self.size / M)
        if self.size >= k:
            return '[%1.1fk]' % (self.size / k)
        return '[%d]' % self.size

    def Get_Message(self):
        return "%s: Random, within %d, count %d" % (self.name,
                    self.size, self.count)

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("tp->random_fn(tp, %d)" % self.size, write)

    def Get_Page_Size(self):
        return self.size

    def Get_Size_Equation(self):
        return "%d" % self.size

    def Get_Abs_Bound(self, max_loop_size):
        return self.size

class LinkedList(AccessType):
    def __init__(self, name, count, store_count, size):
        AccessType.__init__(self, name, count, store_count)
        assert size > 0
        self.size = size

    def Pattern(self):
        return "D"

    #def Info(self):
    #    return '[%d]' % self.size

    def Get_Message(self):
        return "%s: LinkedList: size %d, count %d" % (self.name,
                self.size, self.count)

    def Get_Sequence_Info(self):
        for i in xrange(self.count):
            write = (i < self.store_count)
            yield ("%d" % (rng.randint(0, self.size - 1)), write)

    def Get_Page_Size(self):
        return self.size

    def Get_Size_Equation(self):
        return "%d" % self.size

    def Init_Random(self):
        return True

    def Get_Abs_Bound(self, max_loop_size):
        return self.size

    def Get_Align(self):
        return True # Yes, pointer

class Pattern_Data:
    def __init__(self, name):
        self.name = name
        self.all_of_type = []
        self.prop = 0

class Activity:
    def __init__(self, sulphur, regions):
        self.regions = []
        self.lookup = dict()
        self.use_default = set()
        self.operation = []
        self.prop = 0.0
        self.store_set = set()
        self.special = False

        for rdata in sulphur:
            self.store_set = rdata[ "store_set" ]
            if rdata[ "region_num" ] in regions:
                self.regions.append(rdata)
                for info in rdata[ "source_info" ]:
                    bc_id = info[ "bc_id" ]
                    self.use_default.add(bc_id)
                    if not self.lookup.has_key(bc_id):
                        self.lookup[ bc_id ] = []

                    self.lookup[ bc_id ].append(info[ "pc" ])

                self.prop += rdata[ "proportion" ] 

        assert len(self.regions) != 0

    def Get_Numbers(self):
        return [ rdata[ "region_num" ] for rdata in self.regions ]

    def Get_Names(self):
        x = list(set([ rdata[ "func_name" ] for rdata in self.regions ]))
        x.sort()
        return x

    def Add(self, bc_pc_id_list, code, name, **fields):
        max_factor = 0
        all_pcs = set()
        store_count = 0

        for (bc_id, pc_hint) in bc_pc_id_list:
            assert self.lookup.has_key(bc_id), hex(bc_id)

            if pc_hint == 0:
                self.use_default.remove(bc_id)
                pc_list = self.lookup[ bc_id ]
            else:
                pc_list = [ pc_hint ]
                self.use_default.discard(bc_id)

            for pc in pc_list:
                if (not (pc in all_pcs)) and (pc in self.store_set):
                    store_count += 1

                all_pcs.add(pc)


        if code in ("Linear", "GlobalLinear", "LocalLinear"):
            # Stepping through an array.
            self.operation.append(Linear(name, 
                        len(all_pcs), store_count, max_factor, **fields))

        elif code in ("LocalLinearStruct", "LinearStruct"):
            # Stepping through an array of structures.
            self.operation.append(LinearStruct(name, 
                        len(all_pcs), store_count, max_factor, **fields))

        elif code == "Unrolled":
            # Stepping through an array in an unrolled loop
            self.operation.append(Unrolled(name, 
                        len(all_pcs), store_count, max_factor, **fields))

        elif code == "Static":
            # This could be moved outside of the region as it is a
            # fixed address
            if name == "huh":
                assert store_count == 0
                self.operation.append(Literal(len(all_pcs)))
            else:
                self.operation.append(Static(name, len(all_pcs), store_count))

        elif code == "Local":
            # Same as "Static" but the address is unknown until the
            # region is reached.
            self.operation.append(Local(name, len(all_pcs), store_count))

        elif code == "Random":
            # Could be anywhere and anything.
            self.operation.append(Random(name, 
                        len(all_pcs), store_count, **fields))

        elif code == "LinkedList":
            # Structs, linked by "next", could be anywhere
            self.operation.append(LinkedList(name, 
                        len(all_pcs), store_count, **fields))

        else:
            raise Exception("Unknown code: %s" % code)

    def Special(self, name):
        self.special = True

    def Add_Default(self, code, name, **fields):
        self.Add([ (x, 0) for x in self.use_default ], code, name, **fields)

    def Finish(self):
        pass

    def Find_Operations_Of_Type(self, t):
        return [ x for x in self.operation 
                    if isinstance(x, t)  ]

    def Get_Proportion(self, number):
        if len(self.operation) == 0:
            return 0
        return (self.prop * number) / len(self.operation)

def Main():
    data_received = False
    regions = set()
    program = None
    activity = None
    all_activities = dict()
    line_number = 0

    for line in file("benchmarks/benchnotes.txt"):
        line_number += 1
        if line.startswith("++ "):
            fields = line.split()
            if data_received:
                regions = set()
                data_received = False
                activity.Finish()
                activity = None

            program = fields[ 1 ][ : -1 ]
            regions.add(int(fields[ 2 ]))

        if line.startswith("// "):
            try:
                if not data_received:
                    assert activity == None
                    assert program != None
                    sulphur = Get_Sulphur(program)
                    if sulphur != None:
                        activity = Activity(sulphur, regions)

                        if not all_activities.has_key(program):
                            all_activities[ program ] = []

                        all_activities[ program ].append(activity)
                        data_received = True

                if data_received:
                    fields = line.split()

                    if fields[ 1 ] == "specialpattern":
                        activity.Special(fields[ 2 ])
                    else:
                        assert len(fields) > 3, "Not enough fields"
                        code = fields[ 2 ]
                        name = fields[ 3 ]
                        extra = dict()
                        for i in xrange(4, len(fields), 2):
                            (key, value) = (fields[ i ], fields[ i + 1 ])
                            value = int(value)
                            extra[ key ] = value

                        if fields[ 1 ] == "default":
                            activity.Add_Default(code, name, **extra)
                        else:
                            bc_pc_id_list = []
                            for word in fields[ 1 ].split(','):
                                parts = word.split('@')
                                assert 1 <= len(parts) <= 2
                                bc_id = int(parts[ 0 ], 16)
                                if len(parts) == 1:
                                    bc_pc_id_list.append((bc_id, 0))
                                else:
                                    bc_pc_id_list.append((bc_id, 
                                            int(parts[ 1 ], 16)))

                            activity.Add(bc_pc_id_list, code, name, **extra)
            except:
                print program, regions
                print line_number, line
                raise

    if activity != None:
        activity.Finish()


    access_type_list = []
    access_type_name = dict()
    for (name, t) in globals().iteritems():
        if ((type(t) == type(AccessType))
        and issubclass(t, AccessType) 
        and (t != AccessType)):
            access_type_list.append(t)
            access_type_name[ t ] = name

    seq_access_count = 0
    ro_open_count = 0
    rw_open_count = 0
    codes = file("sulphurdata/codes1.tex", "wt")
    desc = file("sulphurdata/summary.tex", "wt")
    desc.write(
r"\begin{tabular}{|c|ll|p{3.2cm}|}\hline" "\n"
r"\ts Nr &\ts  Program &\ts  Function &\ts  Contents \\" "\n"
r"\hline" "\n")
    tetra_py = file("tetra.py", "wt")
    tetra_c = file("tetra.c", "wt")
    tetra_c.write("""
/* GENERATED CODE */

#include <stdlib.h> /* qsort */
#include <assert.h> /* assert */
#include <string.h> /* memset/memcpy */
#include "tetra.h"
#define MAX_PATTERNS 32

typedef struct Tetra_Pattern_struct {
    ulongint    size, value, space, interval;
    int         read_only, close, during, index, score;
} Tetra_Pattern;

static int Sort_By_Value (const void * a, const void * b)
{
    const Tetra_Pattern * tpa = (const Tetra_Pattern *) a;
    const Tetra_Pattern * tpb = (const Tetra_Pattern *) b;
    if (tpa->score < tpb->score) {
        return -1;
    } else if (tpa->score > tpb->score) {
        return 1;
    } else if (tpa->index < tpb->index) {
        return -1;
    } else if (tpa->index > tpb->index) {
        return 1;
    } else {
        return 0;
    }
}

static void Ready_Patterns(Tetra_Params * tp,
            Tetra_Pattern * pattern_setup,
            Tetra_Pattern * pattern, int count)
{
    void * cb = tp->mem_model_cb;
    ulongint space = 0;
    ulongint i;

    memset(pattern, 0, sizeof(Tetra_Pattern) * MAX_PATTERNS);
    qsort(pattern_setup, count, sizeof(Tetra_Pattern), Sort_By_Value);

    for (i = 0; i < count; i++) {
        Tetra_Pattern * p = &pattern_setup[ i ];
        ulongint test = p->value;
        p->space = space;
        if (p->score < 0) {
            p->value = tp->open_fn(cb, p->value,
                    p->size, &space, p->read_only);
        }
        if (p->value != test) {
            p->close = 1;
            if (!tp->page_flag) {
                p->during = 1;
            }
        }
        assert(0 <= p->index);
        assert(p->index < count);
        memcpy(&pattern[ p->index ], p, sizeof(Tetra_Pattern));
    }
}

/* END INITIAL CODE */

""")


    num_base_pointers = []
    all_activities_list = all_activities.items()
    all_activities_list.sort(key = lambda (program, activity_list):
            (program.lower(), activity_list[ 0 ].Get_Names()[ 0 ],
                        activity_list[ 0 ].Get_Numbers()[ 0 ]))
    assignment = []

    for (program, activity_list) in all_activities_list:
        remainder = 1.0
        for activity in activity_list:
            if activity.special:
                continue

            ncode = program
            for n in activity.Get_Numbers():
                ncode += "_%d" % n

            all_ops = activity.Find_Operations_Of_Type(AccessType)

            tetra_c.write("""
static void %s(Tetra_Params * tp) {
    void * cb = tp->mem_model_cb;
""" % ncode)

            all_ops.sort(key = lambda a: a.Get_Page_Size())

            priority = []
            for (i, a) in enumerate(all_ops):
                if ((a.Pattern() == 'D') and a.Read_Only()):
                    priority.append(i)

            for (i, a) in enumerate(all_ops):
                if ((a.Pattern() == 'S') and a.Read_Only() 
                and not i in priority):
                    priority.append(i)

            for (i, a) in enumerate(all_ops):
                if (a.Read_Only() and not i in priority):
                    priority.append(i)

            for (i, a) in enumerate(all_ops):
                if (not i in priority):
                    priority.append(i)

            for (i, a) in enumerate(all_ops):
                if a.Read_Only():
                    ro_open_count += 1
                else:
                    rw_open_count += 1

            tetra_c.write("""
    /* initialisation */
    ulongint i;
    ulongint read_only_remaining = tp->ro_max;
    Tetra_Pattern pattern_setup[ MAX_PATTERNS ];
    Tetra_Pattern pattern[ MAX_PATTERNS ];
    memset (pattern_setup, 0, sizeof(Tetra_Pattern) * MAX_PATTERNS);
    assert ( MAX_PATTERNS >= %d ) ;
    """ % (max(priority) + 1))

            for i in priority:
                a = all_ops[ i ]
                count = len([ x for x in a.Get_Sequence_Info() ])
                tetra_c.write("""
    /* initialise pattern %d */
    i = %d;
    {
        int read_only = 0;
        ulongint size = 0;
        ulongint loop_size = tp->loop_size;
        ulongint access_count = %d; /* count */
        Tetra_Pattern * p = &pattern_setup[ i ];

        if (%d && read_only_remaining) {    /* a.Read_Only() */
            read_only_remaining --; 
            read_only = 1;
        }
        if (tp->page_flag) {
            size = tp->page_size;
        } else if (tp->tiling_flag && %d) { /* a.Pattern() == 'S' */
            size = tp->tile_size;
        } else {
            size = %s;
        }
        if (!%d) {   /* a.Init_Random() */
            access_count *= loop_size;
        }
        p->score = -tp->worth_copying_fn(cb, size, 
                    access_count, read_only);
        p->size = size;
        p->value = tp->new_fn(tp, 0, %d);
        p->read_only = read_only;
        p->close = 0;
        p->during = 0;
        p->index = i;
    """ % (i, i, count,
                    a.Read_Only(),
                    a.Pattern() == 'S', 
                    a.Get_Size_Equation(),
                    a.Init_Random(),
                    a.Get_Abs_Bound(MAX_LOOP_SIZE)))

                if a.Pattern() == 'S':
                    tetra_c.write("""
        p->interval = tp->tile_size / %d; /* abs(a.step) */
        if (p->interval == 0) p->interval = 1;""" % (
                        max(1, abs(a.step))))

                tetra_c.write("""
    }""")
   
            nr = max(priority) + 1
            tetra_c.write("""
    Ready_Patterns(tp, pattern_setup, pattern, %d);""" % (nr))

            tetra_c.write("""
    /* start */
    for (i = 0; i < tp->loop_size; i++) {
        Tetra_Pattern * p = NULL;
        int newloc;
                """)

            operations = []
            for (i, a) in enumerate(all_ops):
                shift = ''
                if ((a.Pattern() == 'S') and (a.step < 0)):
                    shift += (" + (tp->tiling_flag ?"
                            "(tp->tile_size - %d) : 0)" % (-a.step))

                yes = False
                for (offset, write) in a.Get_Sequence_Info():
                    if write:
                        assert not a.Read_Only()
                        operations.append(
                'tp->write_fn(cb, pattern[ %d ].value + %s%s, "%s");\n' 
                                % (i, offset, shift, str(a.Get_Message())))
                    else:
                        operations.append(
                'tp->read_fn(cb, pattern[ %d ].value + %s%s, "%s");\n' 
                                % (i, offset, shift, str(a.Get_Message())))
                    yes = True
               
                assert yes

            rng.shuffle(operations)
            operations.append("tp->do_tick_fn(tp, TICK_LOOP_OVERHEAD);\n")
            tetra_c.write('\n/* operations */\n')
            tetra_c.write(''.join(operations))

            for (i, a) in enumerate(all_ops):
                allow_newloc = False
                tetra_c.write('\n/* generate pattern %d */\n' % i)
                tetra_c.write('p = &pattern[ %d ];\n' % i)
                if a.Init_Random():
                    # New location for next iteration
                    tetra_c.write('newloc = 1;\n')
                    allow_newloc = True
                elif a.Pattern() == 'S':
                    tetra_c.write('newloc = tp->tiling_flag '
                        '&& ((i % p->interval) == (p->interval - 1));\n')
                    allow_newloc = True
                else:
                    tetra_c.write('newloc = 0;\n')

                if allow_newloc:
                    tetra_c.write("""
        if (newloc) {
            if (p->during) {
                p->value = tp->close_fn(cb, p->value);
            }""")
                    if a.Init_Random():
                        multiplicand = 0
                        tetra_c.write("""
            p->value = tp->new_fn(tp, p->value, %d);
            """ % (a.Get_Abs_Bound(MAX_LOOP_SIZE)))
                    else:
                        assert a.Pattern() == 'S'
                        multiplicand = a.step

                    tetra_c.write("""
            if (p->during) {
                int64_t offset = tp->tiling_flag ? ((int64_t) (i+1) * %d) : 0;
                ulongint space = p->space;
                p->value = tp->open_fn(cb, p->value + offset,
                        p->size, &space, p->read_only) - offset;
            }
        } /* if (newloc) */""" % multiplicand)
                
            tetra_c.write("""
    } /* for */
   
    for (i = 0; i < %d; i++) {
        if (pattern[i].close) {
            tp->close_fn(cb, pattern[i].value);
        }
    }""" % (nr))

            tetra_c.write("""
} /* function */
""")

            pattern_data = dict()
            present_names = []
            total = 0
            for at in access_type_list:
                all_of_type = activity.Find_Operations_Of_Type(at)
                if len(all_of_type) != 0:
                    name = all_of_type[ 0 ].Pattern()
                    if not pattern_data.has_key(name):
                        pattern_data[ name ] = Pattern_Data(name)
                        present_names.append(name)

                    pd = pattern_data[ name ]
                    pd.all_of_type.extend(all_of_type)
                    prop = activity.Get_Proportion(len(all_of_type))
                    pd.prop += prop
                    for a in all_of_type:
                        sz = len([ x for x in a.Get_Sequence_Info() ])
                        total += sz
          
            present_names.sort()

            if total != 0:
                code = []
                func_name = activity.Get_Names()[ 0 ]
                if func_name == "susan_smoothing":
                    codes.write(r'\newcommand{\nghz}{%d}' % 
                                (len(assignment) + 1))

                desc.write(r'\ts %d &\ts %s &\ts %s &\ts ' % (len(assignment) + 1,
                            program, Pre_Proc(func_name)))
                tetra_c.write('/* %s,%s,' % (program, func_name))
                seq_access = False
                count_here = 0
                for name in present_names:
                    pd = pattern_data[ name ]
                    for a in pd.all_of_type:
                        sz = len([ x for x in a.Get_Sequence_Info() ])
                        code.append('%d%s%s%s' % (sz, name, a.Info(),
                            [ "*", "" ][ a.Read_Only() ]))
                        desc.write(' ' + code[ -1 ])
                        tetra_c.write(' ' + code[ -1 ])
                        if a.Pattern() == 'S':
                            seq_access = True
                        count_here += 1

                if seq_access:
                    seq_access_count += 1
                desc.write(r' \\')
                desc.write('\n')
                tetra_c.write('*/\n')
                assignment.append((ncode, len(assignment) + 1,
                        program, func_name, code))
                num_base_pointers.append(count_here)

            tetra_c.write('\n\n')

        #desc.write('%1.4f remains in %s\n\n' % (remainder, program))
  
    function_set = set()
    tetra_c.write("Tetra_Method tetra_methods [] = {\n")
    for ass in assignment:
        ncode = ass[ 0 ]
        tetra_c.write("%s,\n" % ncode)
        function_set.add(ncode)

    tetra_c.write("NULL };\n\n")
    tetra_c.write("const char * tetra_descs [] = {\n")
    for ass in assignment:
        ncode = ass[ 0 ]
        tetra_c.write("\"%s\",\n" % ncode)

    tetra_c.write("NULL };\n\n")

    tetra_py.write('ASSIGNMENT = %s\n\n'
                'FN_LIST = %s\n\n' % (repr(assignment), 
                        repr(list(function_set))))
    codes.write(r'\newcommand{\woqi}{%d}' % (len(assignment)))
    codes.write(r'\newcommand{\msqi}{%d}' % (seq_access_count))
    codes.write(r'\newcommand{\abcd}{%d}' % (ro_open_count))
    codes.write(r'\newcommand{\bcde}{%d}' % (rw_open_count))
    codes.write("\n\n")
    codes.close()
    desc.write(r"\hline\end{tabular}" "\n\n")
    desc.close()
    tetra_c.close()
    tetra_py.close()

    num_base_pointers.sort()
    print 'median base pointer count %d' % (
                num_base_pointers[ len(num_base_pointers) / 2 ])
    print 'maximum base pointer count %d' % max(num_base_pointers)
    print 'mean base pointer count %1.2f' % (
                float(sum(num_base_pointers)) / len(num_base_pointers))




if __name__ == "__main__":
    Main()

