/*
 * ?? 2010 submission "Vanadium"
 * main program
 * Copyright (C) Jack Whitham 2010
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <unistd.h>

#include "tetra.h"
#include "smmu.h"
#include "cache.h"
#include "memlib.h"

#include "u64.hh"

#define UNDEFINED (~((ulongint) 0))

typedef struct TDMA_Data_struct {
    const char *    method_name;
    Tetra_Method    method;
    pid_t           child_pid;
    int             ideal;
    ulongint        event_time;
    ulongint        number;
    ulongint *      tdma_pattern;
    ulongint        tdma_pattern_size;
    ulongint        bus_latency;
    ulongint        bus_width;
    ulongint        ideal_event_time, real_event_time;
} TDMA_Data;


static void TDMA_Txfer(Tetra_Params * tp,
                    ulongint size, ulongint min_cycles)
{
    TDMA_Data *     tdma_data = (TDMA_Data *) tp->user_cb;
    ulongint *      tdma_pattern = tdma_data->tdma_pattern;
    ulongint        tdma_pattern_size = tdma_data->tdma_pattern_size;
    ulongint        my_number = tdma_data->number;
    ulongint        event_time = tdma_data->event_time;
    ulongint        absolute_limit = event_time + (tdma_pattern_size * 2);
    ulongint        slot_start, slot_end, slot_size;

    assert (size > 0);
    assert (size < UINT_MAX);

    /* Ideal mode? */
    if (tdma_data->ideal) {
        event_time += tdma_data->bus_latency +
            ((size + tdma_data->bus_width - 1) / tdma_data->bus_width);
        tdma_data->event_time = event_time;
        return;
    }

    /* Schedule this transfer on the bus */
    while (size != 0) {
        /* Find the next TDMA slot */
        slot_start = slot_end = UNDEFINED;

        while (event_time < absolute_limit) {
            if (tdma_pattern[ event_time % 
                            tdma_pattern_size ] != my_number) {
                if (slot_start == UNDEFINED) {
                    /* Not found start of slot yet */
                } else {
                    /* Found end of slot */
                    slot_end = event_time;
                    break;
                }
            } else {
                if (slot_start == UNDEFINED) {
                    /* Found start of slot */
                    slot_start = event_time;
                }
            }
            event_time ++;
        }
        if (slot_start == UNDEFINED) {
            printf("%s: failed, no suitable slots in TDMA schedule.\n", 
                        tdma_data->method_name);
            fflush(stdout);
            exit(1);
        }
        if (slot_end == UNDEFINED) {
            /* Infinite space available! TDMA schedule is dedicated to
             * one of the methods (this one). */
            slot_end = event_time;
        }
        /* How many bytes can be transferred in this slot? */
        slot_size = slot_end - slot_start;
        if (slot_size <= tdma_data->bus_latency) {
            /* No bytes */
        } else {
            /* Some bytes */
            ulongint max_bytes = (slot_size - 
                    tdma_data->bus_latency) * tdma_data->bus_width;
            if (max_bytes > size) {
                /* There is time left over */
                event_time = slot_start + tdma_data->bus_latency +
                    ((size + tdma_data->bus_width - 1) / 
                            tdma_data->bus_width);
                assert (event_time >= slot_start);
                assert (event_time <= slot_end);
                size = 0;
            } else {
                /* All time is used up */
                size -= max_bytes;
                absolute_limit = event_time + (tdma_pattern_size * 2);
            }
        }
    }
    tdma_data->event_time = event_time;
}

static void TDMA_Tick (Tetra_Params * tp, Tick_Type tt)
{
    TDMA_Data *     tdma_data = (TDMA_Data *) tp->user_cb;

    tdma_data->event_time ++;
}

static void Process_Method(TDMA_Data * tdma_data, ulongint ts)
{
    Tetra_Params *  tp = Init_Params ();

    tp->user_cb = tdma_data;
    tp->smmu_very_deterministic = 1;
    tp->tiling_flag = 1;
    tp->tile_size = ts;
    Init_SMMU(tp, 16, CM_OPTION_B);
    tp->do_txfer_fn = TDMA_Txfer;
    tp->do_tick_fn = TDMA_Tick;
    tdma_data->method(tp);
    tp->flush_fn(tp->mem_model_cb);
    tp->destroy_fn(tp->mem_model_cb);
    free(tp);
}

static TDMA_Data * Initialise_Methods(
            ulongint * method_list, 
            ulongint method_list_size,
            ulongint * tdma_pattern,
            ulongint tdma_pattern_size,
            ulongint bus_latency,
            ulongint bus_width)
{
    TDMA_Data * methods;
    ulongint    i;

    /* Create data structure for methods */
    methods = calloc(method_list_size, sizeof(TDMA_Data));
    for (i = 0; i < method_list_size; i++) {
        methods[ i ].method_name = tetra_descs[ method_list [ i ] ];
        methods[ i ].method = tetra_methods[ method_list [ i ] ];
        methods[ i ].tdma_pattern = tdma_pattern;
        methods[ i ].tdma_pattern_size = tdma_pattern_size;
        methods[ i ].bus_latency = bus_latency;
        methods[ i ].bus_width = bus_width;
        methods[ i ].number = i;
    }
    return methods;
}

#if 0
int main (void)
{
    ulongint        method_list [] = { 10, 15, 20, 25 };
    ulongint        method_list_size = 4;
    ulongint        slice_size = 50;
    ulongint        bus_latency = 7;

    ulongint        tdma_pattern_size = slice_size * method_list_size;
    ulongint *      tdma_pattern;
    ulongint        i;
    TDMA_Data *     methods;

    /* Set up TDMA schedule */
    tdma_pattern = calloc(tdma_pattern_size, sizeof(ulongint));
    for (i = 0; i < tdma_pattern_size; i++) {
        tdma_pattern[ i ] = i / slice_size;
    }
    /*for (i = slice_size / 2; i < slice_size; i++) {
        tdma_pattern[ i ] = 1;
    }*/

    methods = Initialise_Methods(method_list, method_list_size,
            tdma_pattern, tdma_pattern_size,
            bus_latency, 4);

    /* Execute methods */
    for (i = 0; i < method_list_size; i++) {
        methods[ i ].ideal = 1;
        Process_Method(&methods[ i ]);
        methods[ i ].ideal_event_time = methods[ i ].event_time;
        methods[ i ].ideal = 0;
        methods[ i ].event_time = 0;
        Process_Method(&methods[ i ]);
        methods[ i ].real_event_time = methods[ i ].event_time;
        printf("%s: ideal %"FU" tdma %"FU"\n",
                methods[ i ].method_name,
                methods[ i ].ideal_event_time,
                methods[ i ].real_event_time);
    }

    free (methods);
    free (tdma_pattern);
    return 0;
}
#endif

int main (void)
{
    ulongint        tdma_pattern_size = 100;
    ulongint        bus_latency = 7;
    ulongint        bus_width = 4;
    ulongint        method_number = 1;

    ulongint *      tdma_pattern;
    ulongint *      rmatrix;
    ulongint *      ts_matrix;
    ulongint        slot_size;

    rmatrix = calloc(tdma_pattern_size + 1, sizeof(ulongint));
    ts_matrix = calloc(tdma_pattern_size + 1, sizeof(ulongint));
    tdma_pattern = calloc(tdma_pattern_size, sizeof(ulongint));

    for (method_number = 0; 
            tetra_methods[ method_number ] != NULL; method_number++) {

        ulongint last;

        for (slot_size = bus_latency + 1; 
                    slot_size <= tdma_pattern_size; slot_size ++) {

            TDMA_Data * methods;
            ulongint    i, ts, best = UNDEFINED;
            ulongint    best_ts = UNDEFINED;
            ulongint    method_list[ 1 ];

            for (i = 0; i < tdma_pattern_size; i++) {
                tdma_pattern[ i ] = (i <= slot_size) ? 0 : UNDEFINED;
            }
            method_list[ 0 ] = method_number;

            methods = Initialise_Methods(method_list, 1,
                    tdma_pattern, tdma_pattern_size,
                    bus_latency, bus_width);

            for (ts = 32; ts <= 128; ts += 4) {
                methods[ 0 ].ideal = 0;
                methods[ 0 ].event_time = 0;
                Process_Method(&methods[ 0 ], ts);
                if (methods[ 0 ].event_time < best) {
                    best = methods[ 0 ].event_time;
                    best_ts = ts;
                }
            }
            last = rmatrix[ slot_size ] = best;
            ts_matrix[ slot_size ] = best_ts;
            free (methods);
        }

        printf("%s", tetra_descs[ method_number ]);
        for (slot_size = bus_latency + 1; 
                    slot_size <= tdma_pattern_size; slot_size ++) {
            printf(",%f,%"FU, 
                    ((double) rmatrix[ slot_size ]) / ((double) last),
                    ts_matrix[ slot_size ]);
        }
        printf("\n");

    }
    free (tdma_pattern);
    free (rmatrix);
    free (ts_matrix);
    return 0;
}


